/* 
 *    gatos.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <iostream>

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>

#include <gatos/config.h>
#include <gatos/gatos.h>
#include <gatos/freq.h>
#include <gatos/cc.h>
#include <gatos/version.h>

#include <input/input.h>

#define MODULE_NAME "gatos"

/* Default values */
#define DEFAULT_WIDTH 320
#define DEFAULT_HEIGHT 240
#define DEFAULT_FPS 25
#define DEFAULT_FRAME_FORMAT FRAME_FORMAT_YUV12
#define DEFAULT_TUNER_FREQUENCY 0

/* Extern asm functions for fast conversion between color formats */
extern "C" short YUYV_to_YUV9(byte *, short int * const, short width, short height);

extern "C" short YUYV_to_YUV12i(byte *, short int * const, short int * const, short width, short height);
extern "C" short UYVY_to_YUV12i(byte *, short int * const, short int * const, short width, short height);

static bool parse_args(int argc, char * argv[]);
static void timer_handler(int signal);
static void capture();
static void print_info();
static bool setframeparameters(frameformat_t format, int width, int height);
static void setfrequency(double frequency);
static void gentestframe();

// WARNING: code cannot be made reentrant as long as we use a SIGALRM based timer
static int fps;
static double frequency;
static struct itimerval t;
static input_callback_f input_callback;
static VideoFrame frame;
static bool init_ok;
static bool capturing;
static bool verbose;

char module_id[] = MODULE_NAME;

extern "C" void * module_init(int argc, char * argv[])
{
  init_ok = false;
  capturing = false;

  input_callback = 0;

  if(!parse_args(argc, argv)) return 0;

  gatos_setverbosity(0);

  if(gatos_init(GATOS_TUNER|GATOS_READ_BUFFER))
    return 0;

  setframeparameters(frame.format, frame.sx, frame.sy);

  if(frequency != 0) setfrequency(frequency);

  gatos_enable_video(true);
  gatos_enable_sound(true);
  gatos_setmapped(true);
  gatos_enable_capture(true);

  /* Hook to the timer signal */
  signal(SIGALRM, timer_handler);
  t.it_value.tv_sec = 0;
  t.it_value.tv_usec = 1000000/fps;
  t.it_interval.tv_sec = 0;
  t.it_interval.tv_usec = 1000000/fps;
  setitimer(ITIMER_REAL, &t, NULL);

  if(verbose) print_info();
  
  init_ok = true;

  return((void *) 1);
}

extern "C" void module_close(void * context)
{
  if(init_ok)
  {
    signal(SIGALRM, SIG_IGN);

    gatos_enable_capture(false);
    gatos_enable_video(false);
    gatos_enable_sound(false);
    gatos_setmapped(false);
    gatos_terminate();
    init_ok = false;
  }
}

extern "C" void input_set_callback(void * context, input_callback_f callback)
{
  input_callback = callback;
}

bool parse_args(int argc, char * argv[])
{
  frame.sx = DEFAULT_WIDTH;
  frame.sy = DEFAULT_HEIGHT;
  fps = DEFAULT_FPS;
  frequency = DEFAULT_TUNER_FREQUENCY;
  frame.format = DEFAULT_FRAME_FORMAT;
  verbose = false;

  for(register int i = 0; i < argc; i++)
  {
    if(!strncmp(argv[i], "module=", strlen("module=")))
    {
      if(strcmp(argv[i]+strlen("module="), MODULE_NAME)) return(false);
    }
    else
    if(!strcmp(argv[i], "help"))
    {
      fprintf(stderr, "   %s module options:\n", MODULE_NAME);
      fprintf(stderr, "      width=WIDTH         : set the width of the capture to WIDTH\n");
      fprintf(stderr, "      height=HEIGHT       : set the height of the capture to HEIGHT\n");
      fprintf(stderr, "      fps=FPS             : set the number of frames per second to FPS\n");
      fprintf(stderr, "      frequency=FREQUENCY : set the tuner frequency to FREQUENCY\n");
      fprintf(stderr, "      format=FORMAT       : set the frame color format to FORMAT\n");
      fprintf(stderr, "                            FORMAT is one of { yuv12 (default), yuv9, yuyv }\n");
      fprintf(stderr, "      verbose             : set the verbose mode\n");
      fprintf(stderr, "\n");
      return(false);
    }
    else
    if(!strncmp(argv[i], "width=", strlen("width=")))
    {
      frame.sx = atoi(argv[i]+strlen("width="));
    }
    else
    if(!strncmp(argv[i], "height=", strlen("height=")))
    {
      frame.sy = atoi(argv[i]+strlen("height="));
    }
    else
    if(!strncmp(argv[i], "fps=", strlen("fps=")))
    {
      fps = atoi(argv[i]+strlen("fps="));
      if(!fps) fps = DEFAULT_FPS;
    }
    else
    if(!strncmp(argv[i], "frequency=", strlen("frequency=")))
    {
      frequency = atoi(argv[i]+strlen("frequency="));
    }
    else
    if(!strncmp(argv[i], "format=", strlen("format=")))
    {
      char * color_format = argv[i]+strlen("format=");

      if(!strcmp(color_format, "yuv12") || !strcmp(color_format, "YUV12"))
	frame.format = FRAME_FORMAT_YUV12;

      if(!strcmp(color_format, "yuv9") || !strcmp(color_format, "YUV9"))
	frame.format = FRAME_FORMAT_YUV9;

      if(!strcmp(color_format, "yuyv") || !strcmp(color_format, "YUYV"))
	frame.format = FRAME_FORMAT_YUYV;
    }
    else
    if(!strcmp(argv[i], "verbose"))
    {
      verbose = true;
    }
    else
      return(false);
  }
  return(true);
}

void timer_handler(int sig)
{
  /* Handle the timer signal */
  if(sig != SIGALRM) return;

  if(input_callback)
  {
    capture();
    input_callback(&frame);
  }
}

void capture()
{
  volatile byte *buf0=ATIFB+gatos.buffer0,
                *buf1=ATIFB+gatos.buffer1;

  //gatos_enable_video(0);

  // TODO: conversions are too slow (26 ms @ 320x240x25 on my K6-2 350)


  //frame.bitmap = (byte *) buf0;

  //memcpy(frame.bitmap, (void *) buf0, gatos.xcapt*gatos.ycapt);
  if(!frame.bitmap) return;

  switch(frame.format)
  {
    case FRAME_FORMAT_YUV9:
      // BUG: only one of the fields of the picture is read
      YUYV_to_YUV9(frame.bitmap, (short *) buf0, gatos.xcapt, gatos.ycapt);
    break;

    case FRAME_FORMAT_YUV12:
      if(!strncmp("Rage128", gatos.aticard[gatos.cardidx].ident, strlen("Rage128")))
	YUYV_to_YUV12i(frame.bitmap, (short *) buf1, (short *) buf0, gatos.xcapt, gatos.ycapt);
      if(!strncmp("Mach64", gatos.aticard[gatos.cardidx].ident, strlen("Mach64")))
	UYVY_to_YUV12i(frame.bitmap, (short *) buf0, (short *) buf1, gatos.xcapt, gatos.ycapt);
    break;

    case FRAME_FORMAT_YUYV:
      // BUG: This is slow when passing the buffer directly to the encoder!
      //      I can't understand why.
      memcpy(frame.bitmap, (void *) buf0, gatos.xcapt*gatos.ycapt);
      //frame.bitmap = (byte *) buf0;
    break;

    default:
      cerr << "Fatal: unsupported capture format\n" << endl;
      exit(0);
    break;
  }

  //gatos_enable_video(1);
}

void print_info()
{
  cerr << "capture device : ATI " << gatos.aticard[gatos.cardidx].ident << endl;
  cerr << "capture x: " << gatos_xcapt() << " y: " << gatos_ycapt() << endl;
  cerr << "channel : " << gatos_ycapt() << endl; 
  cerr << "frequency range : " << gatos_minfreq() << "->" << gatos_maxfreq() << endl; 
}

bool setframeparameters(frameformat_t format, int width, int height)
{
  frame.sx = width;
  frame.sy = height;
  frame.format = format;

  switch(frame.format)
  {
    case FRAME_FORMAT_YUV9:
      // pixels coded on 9 bits
      frame.bitmap = (byte *) malloc(width*height*9/8);  
    break;

    case FRAME_FORMAT_YUV12:
      // pixels coded on 12 bits
      frame.bitmap = (byte *) malloc(width*height*12/8);
    break;

    case FRAME_FORMAT_YUYV:
      // native format; pixels coded on 16 bits
      frame.bitmap = (byte *) malloc(width*height*16/8);
    break;

    default:
    return(false);
  }

  gatos_setcapturesize(width, height);

  return(true);
}

void setfrequency(double frequency)
{
  gatos_settuner(frequency);
}

void gentestframe()
{
    static int i = 0;
    int x, y;

    // Y    
    for(x = 0; x < frame.sx; x++)
        for(y = 0; y < frame.sy; y++)
                frame.bitmap[y * frame.sx + x] = (x + y + i) & 255;
    // U
    for(x = 0; x < frame.sx/4; x++)
        for(y = 0; y < frame.sy/4; y++)
                frame.bitmap[frame.sx * frame.sy + y * frame.sx / 4 + x] = (x + y - i) & 255;
                
    // V
    for(x = 0; x < frame.sx/4; x++)
        for(y = 0; y < frame.sy/4; y++)
                frame.bitmap[17 * frame.sx * frame.sy / 16 + y * frame.sx / 4 + x] = (x - y + i) & 255;

    i++;
}

