/* 
 *    v4l.cpp
 *
 *	Copyright (C) Damien Chavarria - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <signal.h>

#include <input/input.h>
#include <input/bgrab.h>

#define MODULE_NAME "v4l"

/* Default values */
#define DEFAULT_WIDTH 320
#define DEFAULT_HEIGHT 240
#define DEFAULT_FPS 25
#define DEFAULT_FRAME_FORMAT FRAME_FORMAT_YUV12
#define DEFAULT_TUNER_FREQUENCY 0
#define DEFAULT_TUNER_MODE autodetect
#define DEFAULT_CHANNEL 0

enum tunermode_t { autodetect, pal, secam, ntsc };

static bool parse_args(int argc, char * argv[]);
static void timer_handler(int signal);
static void capture();
static void print_info();
static bool set_frame_parameters(frameformat_t format, int width, int height);
static void set_frequency(double frequency);
static void set_tuner_mode(tunermode_t mode,int channel);

// WARNING: code cannot be made reentrant as long as we use a SIGALRM based timer
static int fps;
static double frequency;
static tunermode_t mode;
static struct itimerval t;
static input_callback_f input_callback;
static VideoFrame frame;
static struct fgdevice fg;
static int channel;
static bool verbose;

char module_id[] = MODULE_NAME;

extern "C" void * module_init(int argc, char * argv[])
{
  input_callback = 0;

  if(!parse_args(argc, argv)) return(0);

  if (fg_open_device (&fg, "/dev/video") != 0)
    return(0);

  set_frame_parameters(frame.format, frame.sx, frame.sy);

  set_tuner_mode(mode,channel);

  if(frequency != 0) set_frequency(frequency);

  /* Hook to the timer signal */
  signal(SIGALRM, timer_handler);

  t.it_value.tv_sec = 0;
  t.it_value.tv_usec = 1000000/fps;
  t.it_interval.tv_sec = 0;
  t.it_interval.tv_usec = 1000000/fps;
  setitimer(ITIMER_REAL, &t, NULL);

  if(verbose) print_info();

  return((void *) 1);
}

extern "C" void input_set_callback(void * context, input_callback_f callback)
{
  input_callback = callback;
}

bool parse_args(int argc, char * argv[])
{
  frame.sx = DEFAULT_WIDTH;
  frame.sy = DEFAULT_HEIGHT;
  fps = DEFAULT_FPS;
  frequency = DEFAULT_TUNER_FREQUENCY;
  mode = DEFAULT_TUNER_MODE;
  frame.format = DEFAULT_FRAME_FORMAT;
  channel = DEFAULT_CHANNEL;
  verbose = false;

  for(register int i = 0; i < argc; i++)
  {
    if(!strncmp(argv[i], "module=", strlen("module=")))
    {
      if(strcmp(argv[i]+strlen("module="), MODULE_NAME)) return(false);
    }
    else
    if(!strcmp(argv[i], "help"))
    {
      fprintf(stderr, "   %s module options:\n", MODULE_NAME);
      fprintf(stderr, "      width=WIDTH         : set the width of the capture to WIDTH\n");
      fprintf(stderr, "      height=HEIGHT       : set the height of the capture to HEIGHT\n");
      fprintf(stderr, "      fps=FPS             : set the number of frames per second to FPS\n");
      fprintf(stderr, "      frequency=FREQUENCY : set the tuner frequency to FREQUENCY\n");
      fprintf(stderr, "      channel=CHANNEL     : set the tuner channel to CHANNEL\n");
      fprintf(stderr, "      mode=MODE           : set the tuner mode to MODE\n");
      fprintf(stderr, "                            MODE is one of { auto (default), pal, secam, ntsc }\n");
      fprintf(stderr, "      format=FORMAT       : set the frame color format to FORMAT\n");
      fprintf(stderr, "                            FORMAT is one of { yuv12 (default), yuv9, yuyv }\n");
      fprintf(stderr, "      verbose             : set the verbose mode\n");
      fprintf(stderr, "\n");
      return(false);
    }
    else
    if(!strncmp(argv[i], "width=", strlen("width=")))
    {
      frame.sx = atoi(argv[i]+strlen("width="));
    }
    else
    if(!strncmp(argv[i], "height=", strlen("height=")))
    {
      frame.sy = atoi(argv[i]+strlen("height="));
    }
    else
    if(!strncmp(argv[i], "fps=", strlen("fps=")))
    {
      fps = atoi(argv[i]+strlen("fps="));
      if(!fps) fps = DEFAULT_FPS;
    }
    else
    if(!strncmp(argv[i], "frequency=", strlen("frequency=")))
    {
      frequency = atoi(argv[i]+strlen("frequency="));
    }
    else
    if(!strncmp(argv[i], "mode=", strlen("mode=")))
    {
      frequency = atoi(argv[i]+strlen("mode="));
    }
    else
    if(!strncmp(argv[i], "format=", strlen("format=")))
    {
      char * color_format = argv[i]+strlen("format=");

      if(!strcmp(color_format, "yuv12") || !strcmp(color_format, "YUV12"))
	frame.format = FRAME_FORMAT_YUV12;

      if(!strcmp(color_format, "yuv9") || !strcmp(color_format, "YUV9"))
	frame.format = FRAME_FORMAT_YUV9;
    }
    else
    if(!strncmp(argv[i], "channel=", strlen("channel=")))
    {
      fprintf(stderr,"CHANNEL=%s\n",argv[i]);
      channel = atoi(argv[i]+strlen("channel="));
    }
    else
    if(!strcmp(argv[i], "verbose"))
    {
      verbose = true;
    }
    else
      return(false);
  }
  return(true);
}

void timer_handler(int signal)
{
  /* Handle the timer signal */

  if(signal != SIGALRM) return;

  if(input_callback)
  {
    capture();
    input_callback(&frame);
  }
}

void set_tuner_mode(tunermode_t mode,int channel)
{
 /* Set the tuner mode */
 if(verbose) fprintf(stderr,"channel: %d\n",channel);
 switch(mode)
 {
   case secam:
     fg_set_channel (&fg,channel,VIDEOMODE_SECAM);
   break;

   case pal:
     fg_set_channel (&fg,channel,VIDEOMODE_PAL);
   break;

   case ntsc:
     fg_set_channel (&fg,channel,VIDEOMODE_NTSC);
   break;
   default:
     fg_set_channel (&fg,channel,VIDEOMODE_AUTO);
   break;
 }
}

bool set_frame_parameters(frameformat_t format, int width, int height)
{
 frame.format = format;
 frame.sx = width;
 frame.sy = height;

 switch(format)
 {
   case FRAME_FORMAT_YUV12:
     fg_start_grab_image(&fg, width, height, FORMAT_YUV422P);
     frame.bitmap = (byte *) malloc(width*height*12/8);
   break;
   case FRAME_FORMAT_YUV9:
     fg_start_grab_image(&fg, width, height, FORMAT_YUV411P);
     frame.bitmap = 0;
   break;
   default:
   break;
 }

 return(true);
}

void capture()
{
  int x, y;
  byte *image;

  fg_get_next_image(&fg);

  switch(frame.format)
    {
      case  FRAME_FORMAT_YUV12:

	memcpy(frame.bitmap, fg.current_image, frame.sx*frame.sy);
	
	image = (byte *) fg.current_image;

	for (y=0; y<frame.sy/2; y++)
	  for (x=0; x<frame.sx/2; x++)
	    {
	      frame.bitmap[frame.sx*frame.sy + y*frame.sx/2 + x] =  image[frame.sx*frame.sy + frame.sx*frame.sy/2+  y*frame.sx + x];
	      frame.bitmap[frame.sx*frame.sy +frame.sx*frame.sy/4 + y*frame.sx/2 +x] = image[frame.sx*frame.sy + y*frame.sx + x];
	    }
	break;
    case FRAME_FORMAT_YUV9:

      frame.bitmap = (byte *)fg.current_image;
      break;

    default:
      break;
    }
}

void print_info()
{
 fg_print_info (&fg);
}

void set_frequency(double freq)
{
  unsigned long v4lfreq;

  v4lfreq = (unsigned long) freq / 16;

  if (-1 == ioctl(fg.video_dev, VIDIOCSFREQ, &v4lfreq))
    perror("ioctl VIDIOCSFREQ");
}

