/* 
 *    video.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include <video/video.h>
#include <video/mmx_coder.h>

// Assembly functions
//extern "C" void __CDECL DCT_TEST(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTY_YUV9(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTC_YUV9(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTY_YUV12(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTC_YUV12(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTY_YUYV(Block, unsigned char const *, short int sx);
extern "C" void __CDECL DCTC_YUYV(Block, unsigned char const *, short int sx);
extern "C" void __CDECL QUANTIZE(Block, signed short int *);
extern "C" void __CDECL MMX_END(void);
extern "C" bool __CDECL MMX_DETECT(void);
extern "C" void __CDECL ENCODEYDC(Bitbuffer *, short int, long int *);
extern "C" void __CDECL ENCODECDC(Bitbuffer *, short int, long int *);
extern "C" void __CDECL RLEHUFFZIG(Bitbuffer *, Block);

/* The quatization matrix */
static const int qtable[64] = {
      8,16,19,22,26,27,29,34,
     16,16,22,24,27,29,34,37,
     19,22,26,27,29,34,34,38,
     22,22,26,27,29,34,37,40,
     22,26,27,29,32,35,40,48,
     26,27,29,32,35,40,48,58,
     26,27,29,34,38,46,56,69,
     27,29,35,38,46,56,69,83  };

extern "C" MMXMpegVideoCoder * module_init(int argc, char * argv[])
{
  int16 qscale;
  dword slice_length;
  dword number_of_frames;
  dword start_code_frequency;
  bool verbose;

  qscale = DEFAULT_QSCALE;
  slice_length = DEFAULT_SLICE_LENGTH;
  number_of_frames = DEFAULT_NUMBER_OF_FRAMES;
  start_code_frequency = DEFAULT_START_CODE_FREQUENCY;
  verbose = false;

  for(register int i = 0; i < argc; i++)
  {
    if(!strncmp(argv[i], "module=", strlen("module=")))
    {
      if(strcmp(argv[i]+strlen("module="), MODULE_NAME)) return(0);
    }
    else
    if(!strcmp(argv[i], "help"))
    {
      fprintf(stderr, "   %s module options:\n", MODULE_NAME);
      fprintf(stderr, "      quality=QSCALE      : set the quality of the encoding to QSCALE\n");
      fprintf(stderr, "                            ranges from 3 (best) to 31 (worst)\n");
      fprintf(stderr, "      slice=SLICE_LENGTH  : set the size of output chunks to SLICE_LENGTH\n");
      fprintf(stderr, "      number=NUMBER       : set the maximum number of encoded frames\n");
      fprintf(stderr, "      newsequence=N       : set the number of frames between two sequence head codes\n");
      fprintf(stderr, "      verbose             : set the verbose mode\n");
      fprintf(stderr, "\n");
      return(0);
    }
    else
    if(!strncmp(argv[i], "quality=", strlen("quality=")))
    {
      qscale = atoi(argv[i]+strlen("quality="));
    }
    else
    if(!strncmp(argv[i], "slice=", strlen("slice=")))
    {
      slice_length = atoi(argv[i]+strlen("slice="));
    }
    else
    if(!strncmp(argv[i], "number=", strlen("number=")))
    {
      number_of_frames = atoi(argv[i]+strlen("number="));
    }
    else
    if(!strncmp(argv[i], "newsequence=", strlen("newsequence=")))
    {
      start_code_frequency = atoi(argv[i]+strlen("newsequence="));
    }
    else
    if(!strcmp(argv[i], "verbose"))
    {
      verbose = true;
    }
  }

  /* Check for MMX */
  if(!MMX_DETECT())
  {
    cerr << "MMX not detected." << endl;
    return(0);
  }

  return(new MMXMpegVideoCoder(qscale, slice_length, number_of_frames, start_code_frequency));
}

extern "C" void module_close(MMXMpegVideoCoder * coder)
{
  if(coder) delete coder;
}

extern "C" void video_encode_frame(MMXMpegVideoCoder * coder, VideoFrame * frame)
{
  if(coder) coder->EncodeFrame(*frame);
}

extern "C" void video_set_callback(MMXMpegVideoCoder * coder, video_callback_f callback)
{
  if(coder) coder->SetCallback(callback);
}

void GenSequenceHeader(Bitbuffer * buffer, dword sx, dword sy, int bitrate)
{
  buffer->reset();
  buffer->write(SEQUENCE_HEADCODE, 32); // Head code
  buffer->write(sx, 12);                // Picture size x
  buffer->write(sy, 12);                // Picture size y
  buffer->write(1,   4);                // Pixel aspect ratio
  buffer->write(3,   4);                // Frames per second 5 -> 30
                                           //                   4 -> 29.9
                                           //                   3 -> 25
                                           //                   2 -> 24
                                           //                   1 -> 23.97
   
  buffer->write(bitrate, 18);  // Bit rate (-1 = variable) 102400 bps
  buffer->write(1, 1);         // Marker
  buffer->write(sx*sy, 10);    // Buffer size (arbitrary set to sx*sy)
  buffer->write(0, 1);         // constrained mode
  buffer->write(0, 1);         // load intra matrix
  buffer->write(0, 1);         // load non-intra matrix
}

void GenSequenceEnder(Bitbuffer * buffer)
{
  buffer->reset();
  buffer->write(SEQUENCE_ENDCODE, 32);  // End sequence code
}


static void GenGOPHeader(Bitbuffer * buffer, 
			 int hour, int minute, int second, int frame)
{
    buffer->write(GOP_STARTCODE, 32);            // Gop start code
    buffer->write(0, 1);                         // drop frame flag
    buffer->write(hour, 5);                      // Timecount hours
    buffer->write(minute, 6);                    // Timecount minutes
    buffer->write(1, 1);                         // Marker
    buffer->write(second, 6);                    // Timecount seconds 
    buffer->write(frame, 6);                     // Timecount frames
    buffer->write(1, 1);                         // closed GOP
    buffer->write(0, 1);                         // Broken link

    buffer->write(0, 5);                         // Fill out to 8 bytes
}

static void GenPictureHeader(Bitbuffer * buffer, dword frame_number)
{
    buffer->write(PICT_STARTCODE, 32);
    buffer->write(frame_number, 10);
    buffer->write(I_FRAME,  3);
    buffer->write(0xFFFF, 16);
    buffer->write(0, 1);

    buffer->write(0, 2);                       // Fill out to 8 bytes
}

static void GenSliceHeader(Bitbuffer * buffer, dword vPos, dword qscale)
{
    buffer->write(SLICE_BASE_CODE + vPos, 32); // New slice starting at vpos
    buffer->write(qscale, 5);                  // Quantization scale
    buffer->write(0, 1);                       // That's all folks
}

MMXMpegVideoCoder::MMXMpegVideoCoder(int16 qscale_, dword slice_length_, dword number_of_frames_, dword start_code_frequency_)
{
    register int i;

    qscale = qscale_;
    slice_length = slice_length_;
    number_of_frames = number_of_frames_;
    start_code_frequency = start_code_frequency_;
    /* WARNING: encoded frame size should not exceed 8*slice_length */
    buffer = new Bitbuffer(8*slice_length);
    frame_number = 0;
    callback = 0;

    /* Compute the quantizatin matrix */
    /* This is for I-block: the DC coefficient is never touched */
    qmatrix[0] = (short int) (0x20000L / qtable[0]);
    for(i = 1; i < 64; i++)
      qmatrix[i] = (short int) ((0x100000L+qscale*qtable[i])/(qscale*qtable[i]));

    y_dc_pred = 128;
    cr_dc_pred = cb_dc_pred = 0;

    /* Only MMX capable processors supported currently */
    if(!MMX_DETECT())
      throw;
}

MMXMpegVideoCoder::~MMXMpegVideoCoder()
{
  EndSequence();
  delete buffer;
}

void MMXMpegVideoCoder::SetCallback(video_callback_f callback_)
{
  callback = callback_;
}

void MMXMpegVideoCoder::StartSequence(word sx, word sy)
{
  /* WARNING : bitrate cannot be determined and is set arbitrarly */
  /*           anyway, most decoders don't care about the bitrate */
  /*           specified in the sequence header                   */
    GenSequenceHeader(buffer, sx, sy, 4000);
    if(callback) callback(buffer->data, (buffer->nbits + 7) / 8);
}

void MMXMpegVideoCoder::EndSequence()
{

  /* This code was meant to correct bitrate after the sequence has been  */
  /* written. However, this cannot work on network streams (cannot seek) */

/*
  int32 bitrate;
  int32 time;
  streampos current_pos;
  
  time = (timer->GetCount() - time_zero);

  if(time)
  {
    bitrate = out.tellp() - stream_start;
    time *= 50;
    time /= timer->GetBase(); 
    bitrate /= time;
  }
  else
  {
    bitrate = 0;
    debug.OutputString("Bitrate is null! \n");
  }   
    
  current_pos = out.tellp();
  out.seekp(stream_start);
  WriteSequenceHeader(out, sx, sy, bitrate);
  out.seekp(current_pos);
*/

    GenSequenceEnder(buffer);
    if(callback) callback(buffer->data, (buffer->nbits + 7) / 8);
}

int32 MMXMpegVideoCoder::EncodeFrame(VideoFrame const & frame)
{
    int x, y;
    register unsigned char const * input_y;
    register unsigned char const * input_cr;
    register unsigned char const * input_cb;
    int line_y, line_c;
    int incx_y, incx_c;
    int incy_y, incy_c;
    void (* DCTY)(Block, unsigned char const *, short int sx);
    void (* DCTC)(Block, unsigned char const *, short int sx);
    int currenttime = 0;

    buffer->reset();

    /* Send a new sequence start code once every 'sequence_length' frames */
    if (!(frame_number % start_code_frequency))
      StartSequence(frame.sx, frame.sy);

    GenGOPHeader(buffer, currenttime / 3600000,
		         currenttime / 60000,
		         currenttime / 1000,
		         currenttime / 40);

    GenPictureHeader(buffer, 0);

    GenSliceHeader(buffer, 1, qscale);
    
    y_dc_pred = 128;
    cr_dc_pred = cb_dc_pred = 0;
    
    switch(frame.format)
    {
      case FRAME_FORMAT_YUV9:
	incx_y = DCTSIZE*2;
	incx_c = DCTSIZE/2;
	incy_y = DCTSIZE*2;
	incy_c = DCTSIZE/2;
	line_c = frame.sx/4;
	line_y = frame.sx;
	input_y = frame.bitmap;
	input_cr = frame.bitmap + frame.sx * frame.sy;
	input_cb = frame.bitmap + frame.sx * frame.sy + frame.sx * frame.sy/16;
	DCTY = DCTY_YUV9;
	DCTC = DCTC_YUV9;
      break;

      case FRAME_FORMAT_YUV12:
	incx_y = DCTSIZE*2;
	incx_c = DCTSIZE;
	incy_y = DCTSIZE*2;
	incy_c = DCTSIZE;
	line_c = frame.sx/2;
	line_y = frame.sx;
	input_y = frame.bitmap;
	input_cr = frame.bitmap + frame.sx * frame.sy;
	input_cb = frame.bitmap + frame.sx * frame.sy + frame.sx * frame.sy/4;
	DCTY = DCTY_YUV12;
	DCTC = DCTC_YUV12;
      break;

      case FRAME_FORMAT_YUYV:
	incx_y = 2*DCTSIZE*2;
	incx_c = 2*DCTSIZE;
	incy_y = DCTSIZE*2;
	incy_c = DCTSIZE;
	line_c = 4*frame.sx;
	line_y = 2*frame.sx;
	input_y = frame.bitmap;
	input_cr = frame.bitmap;
	input_cb = frame.bitmap;
       	DCTY = DCTY_YUYV;
       	DCTC = DCTC_YUYV;
      break;
      
      default:
      return(0);
    }

    for (y = 0; y < frame.sy / DCTSIZE; y += 2) 
    {
      // Generate a new slice when more than slice_length bytes are generated (aligned on macro block)
      if((!(y & 2)) && buffer->nbits > slice_length*8)
      {
	if(callback) callback(buffer->data, (buffer->nbits + 7) / 8);
	buffer->reset();
	GenSliceHeader(buffer, 1 + (y>>1), qscale);
	
	y_dc_pred = 128;
	cr_dc_pred = cb_dc_pred = 0;
      }

      for (x = 0; x < frame.sx / DCTSIZE; x += 2)
      {
	// Generate addr incr code.
	buffer->write(1, 1);

	// Generate mb type code.
	buffer->write(1, 1); 

	// Y blocks
        DCTY(block, input_y, line_y);
	QUANTIZE(block, qmatrix);
	ENCODEYDC(buffer, block[0][0], &y_dc_pred);
	RLEHUFFZIG(buffer, block);

	DCTY(block, input_y + (incx_y >> 1), line_y);
	QUANTIZE(block, qmatrix);
	ENCODEYDC(buffer, block[0][0], &y_dc_pred);
	RLEHUFFZIG(buffer, block);

	DCTY(block, input_y + (line_y << LG2_DCTSIZE), line_y);
	QUANTIZE(block, qmatrix);
	ENCODEYDC(buffer, block[0][0], &y_dc_pred);
	RLEHUFFZIG(buffer, block);

	DCTY(block, input_y + (line_y << LG2_DCTSIZE) + (incx_y >> 1), line_y);
    	QUANTIZE(block, qmatrix);
       	ENCODEYDC(buffer, block[0][0], &y_dc_pred);
       	RLEHUFFZIG(buffer, block);

	// CB block
       	DCTC(block, input_cb, line_c);
	QUANTIZE(block, qmatrix);
       	ENCODECDC(buffer, block[0][0], &cb_dc_pred);
       	RLEHUFFZIG(buffer, block);
	
	// CR block
	DCTC(block, input_cr, line_c);
       	QUANTIZE(block, qmatrix);
       	ENCODECDC(buffer, block[0][0], &cr_dc_pred);
       	RLEHUFFZIG(buffer, block);

	input_y  += incx_y;
	input_cb += incx_c;
	input_cr += incx_c;
      }
      input_y +=  (incy_y - 1) * line_y;
      input_cb += (incy_c - 1) * line_c;
      input_cr += (incy_c - 1) * line_c;
    }

    MMX_END();

    if(callback) callback(buffer->data, (buffer->nbits + 7) / 8);

    buffer->reset();

    frame_number++;

    /* Stop after 'number_of_frames' have been encoded */
    if(number_of_frames && frame_number == number_of_frames)
      raise(SIGINT);

    return(1);
}
